#ifndef ATOMIUM
#include <stdio.h>
#include <stdlib.h>
#else
#include <atomium.h>
#endif

/* CIF */
#define MBROWS 18
#define MBCOLS 22

/* Level 3 functions */

int
wrap (int coord, int max) {
   if (coord < 0) return 0;
   if (coord >= max) return max-1;
   return coord;
}

unsigned char
clip(int value) {
   if (value < 0) return (char) 0;
   if (value > 255) return (char) 255;
   return (unsigned char) value;
}

unsigned char
convolV0(unsigned char value[MBROWS*8][MBCOLS*8],
      int col,
      int start_row,
      int stop_row) {
   int i=0, sum=128;
   const int coeffs[] = { 3, -16, 67, 227, -32, 7};
   for (; start_row+i <= stop_row;i++) {
      sum += value[wrap(start_row+i, MBROWS*8)][col]*coeffs[i];
   }
   return clip(sum>>8);
}

unsigned char
convolV1(unsigned char value[MBROWS*8][MBCOLS*8],
      int col,
      int start_row,
      int stop_row) {
   int i=0, sum=128;
   const int coeffs[] = {7, -32, 227, 67, -16, 3};
   for (; start_row+i <= stop_row;i++) {
      sum += value[wrap(start_row+i, MBROWS*8)][col]*coeffs[i];
   }
   return clip(sum>>8);
}

unsigned char
convolH0(unsigned char value[MBROWS*16][MBCOLS*8],
      int row,
      int start_col,
      int stop_col) {
   int i=0, sum=128;
   const int coeffs[] = {5, -21, 70, 228, -37, 11};
   for (;start_col+i <= stop_col;i++) {
      sum += value[row][wrap(start_col+i, MBCOLS*8)]*coeffs[i];
   }
   return clip(sum>>8);
}

unsigned char
convolH1(unsigned char value[MBROWS*16][MBCOLS*8],
      int row,
      int start_col,
      int stop_col) {
   int i=0, sum=128;
   const int coeffs[] = {11, -37, 228, 70, -21, 5};
   for (; start_col+i <= stop_col;i++) {
      sum += value[row][wrap(start_col+i, MBCOLS*8)]*coeffs[i];
   }
   return clip(sum>>8);
}

/*   {104597, 132201, 25675, 53279} ITU-R Rec. 624-4 System B, G */
/*   crv   , cbu   , cgu  , cgv                         */

unsigned char
ycrcb2r ( int y, int cr, int cb ) {
   const int crv = 104597; /* 1.596 * 65536 */
   y = 76309 * (y - 16); /* (255/219)*65536 */
   return  clip((y + crv*(cb-128) + 32768)>>16);
}

unsigned char
ycrcb2g ( int y, int cr, int cb ) {
   const int cgu = 25675; /* 0.392 * 65536 */
   const int cgv = 53279; /* 0.813 * 65536 */
   y = 76309 * (y - 16); /* (255/219)*65536 */
   return  clip((y - cgu*(cr-128) - cgv*(cb-128) + 32768)>>16);
}

unsigned char
ycrcb2b ( int y, int cr, int cb ) {
   const int cbu = 132201; /* 2.017 *65536 */
   y = 76309 * (y - 16); /* (255/219)*65536 */
   return  clip((y + cbu*(cr-128) + 32768)>>16);
}

/* Level 2/1 functions */

int
main (int argvc, char* argv) {

unsigned char     Y[MBROWS*16][MBCOLS*16];
unsigned char Cr420[MBROWS*8][MBCOLS*8];
unsigned char Cb420[MBROWS*8][MBCOLS*8];

unsigned char Cr422[MBROWS*16][MBCOLS*8];
unsigned char Cb422[MBROWS*16][MBCOLS*8];

int row,col;

unsigned char Cr444[MBROWS*16][MBCOLS*16];
unsigned char Cb444[MBROWS*16][MBCOLS*16];

unsigned char RGB[MBROWS*16][3*MBCOLS*16];

FILE* fp;

if ((fp = fopen("akicif.y", "rb")) != NULL) {
   fread(Y, sizeof(unsigned char), (MBROWS*16)*(MBCOLS*16), fp);
   fclose(fp);
}

if ((fp = fopen("akicif.u", "rb")) != NULL) {
   fread(Cr420, sizeof(unsigned char), (MBROWS*8)*(MBCOLS*8), fp);
   fclose(fp);
}

if ((fp = fopen("akicif.v", "rb")) != NULL) {
   fread(Cb420, sizeof(unsigned char), (MBROWS*8)*(MBCOLS*8), fp);
   fclose(fp);
}

for (row=0; row < MBROWS*8; row++) {
   for (col=0; col < MBCOLS*8; col++) {
      Cr422[2*row  ][col] = convolV0(Cr420, col, row-3, row+2);
      Cr422[2*row+1][col] = convolV1(Cr420, col, row-2, row+3);
   }
}

for (row=0; row < MBROWS*8; row++) {
   for (col=0; col < MBCOLS*8; col++) {
      Cb422[2*row  ][col] = convolV0(Cb420, col, row-3, row+2);
      Cb422[2*row+1][col] = convolV1(Cb420, col, row-2, row+3);
   }
}

for (row=0; row < MBROWS*16; row++) {
   for (col=0; col < MBCOLS*8; col++) {
      Cr444[row][2*col  ] = convolH0(Cr422, row,col-3,col+2);
      Cr444[row][2*col+1] = convolH1(Cr422, row,col-2,col+3);
   }
}

for (row=0; row < MBROWS*16; row++) {
   for (col=0; col < MBCOLS*8; col++) {
      Cb444[row][2*col  ] = convolH0(Cb422, row,col-3,col+2);
      Cb444[row][2*col+1] = convolH1(Cb422, row,col-2,col+3);
   }
}

for (row=0; row < MBROWS*16; row++) {
   for (col=0; col < MBCOLS*16; col++) {
      RGB[row][3*col+0] = ycrcb2r(Y[row][col], Cr444[row][col], Cb444[row][col]);
      RGB[row][3*col+1] = ycrcb2g(Y[row][col], Cr444[row][col], Cb444[row][col]);
      RGB[row][3*col+2] = ycrcb2b(Y[row][col], Cr444[row][col], Cb444[row][col]);
   }
}

if ((fp = fopen("out.ppm", "wb")) != NULL) {
   fprintf(fp, "%c%c\n%d %d\n%d\n", 'P', '6', MBCOLS*16, MBROWS*16, 255);
   fwrite(RGB, sizeof(unsigned char), (MBROWS*16)*(3*MBCOLS*16), fp);
   fclose(fp);
}

return 0;
}
